/* NoSuch MIDI - Copyright 1997-1998 by Tim Thompson */

package com.nosuch.cooking;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import java.io.*;
import java.net.*;
import com.nosuch.midi.*;

public class BoomerState implements Externalizable {

	int nsteps;
	int startQuant;
	int stepsize;
	int cellLimit;
	boolean gradual;
	boolean flash;
	Vector rows;	// a Vector of DrumRow
	Vector drums;	// a Vector of DrumInfo, for available choices

	private Boomer parent;

	public BoomerState() {
		startQuant = MidiSystem.clicksPerBeat();
		nsteps = 0;
		rows = new Vector();
		drums = new Vector();

		// fill in drums vector with list of Drums
		DrumMap dm = new DrumMap();
		for ( int nd = 0 ; nd<dm.size(); nd++ ) {
			drums.addElement(new DrumInfo(dm.getName(nd), dm.getPhrase(nd), dm.getCategory(nd) ));
		}

		stepsize = MidiSystem.clicksPerBeat()/4;
		cellLimit = 4;
		gradual = false;
		flash = true;
	}

	public void setParent(Boomer p) {
		parent = p;
	}

	public void writeExternal(ObjectOutput out) throws IOException {
		out.writeInt(nsteps);
		out.writeInt(startQuant);
		out.writeInt(stepsize);
		out.writeInt(cellLimit);
		out.writeBoolean(gradual);
		out.writeBoolean(flash);
		out.writeInt(rows.size());
		for ( int n=0; n<rows.size(); n++ ) {
			out.writeObject((DrumRow)rows.elementAt(n));
		}
		out.writeObject(drums);
	}
	public void readExternal(ObjectInput in) throws IOException {
		nsteps = in.readInt();
		startQuant = in.readInt();
		stepsize = in.readInt();
		cellLimit = in.readInt();
		gradual = in.readBoolean();
		flash = in.readBoolean();

		int nrows = in.readInt();
		rows = new Vector();
		try {
			for ( int n=0; n<nrows; n++ ) {
				DrumRow dr = (DrumRow)in.readObject();
				rows.addElement(dr);
			}
			drums = (Vector)in.readObject();
		} catch (ClassNotFoundException e) {
			System.out.println("Error reading external into BoomerState");
		}
	}

	void initFromPhrase(MidiPhrase p) {

		p = (MidiPhrase) p.clone();	// so we can mess with it

		p.setDuration(stepsize);
		p.setLength(Alg.latest(p));

		Vector notes = new Vector();

		// Find the unique notes (matching only pitch/channel)

		boolean found;
		for ( Enumeration e=p.getEnumeration(); e.hasMoreElements(); ){
			MidiMsg m = (MidiMsg) e.nextElement();

			found = false;
			for ( Enumeration e2=notes.elements(); e2.hasMoreElements(); ) {
				MidiMsg m2 = (MidiMsg)e2.nextElement();
				if ( m2.getPitch() == m.getPitch() && m2.getChannel() == m.getChannel() ) {
					found = true;
					break;
				}
			}
			if ( found == false ) {
				notes.addElement(m);
			}
		}
		if ( notes.size() <= 0 ) {
			System.out.println("Not enough notes in phrase!?");
			return;
		}

		// now create a row for each unique note

		setNrows(notes.size());
		int rn = 0;
		for ( Enumeration e=notes.elements(); e.hasMoreElements(); ){
			MidiMsg m = (MidiMsg) e.nextElement();
			int pit = m.getPitch();
			int ch = m.getChannel();
			if ( pit > 0 ) {

				// Look up pit,ch in drums
				String dname = getNameOf(pit,ch,drums);
				if ( dname != null ) {
					changeDrum(rn,dname);
				}
				else {
					MidiPhrase np = new MidiPhrase();
					np.merge(m);
					np.setTime(0);
					dname = np.toString();
					DrumInfo di = new DrumInfo(dname, np, "other" );
					drums.addElement(di);
					changeDrum(rn,dname);
				}
			}
			rn++;
		}

		parent.gui.resetDrumMenu();

		// Now fill in the cells that should be filled in order
		// to represent the given phrase.

		long lng = p.getLength();
		if ( (lng%(4*stepsize)) != 0 ) {
			lng = Alg.nextquant(lng,4*stepsize);
		}
		int nsteps = (int)(lng / stepsize);
		setNsteps(nsteps);

		for ( Enumeration e=p.getEnumeration(); e.hasMoreElements(); ){
			MidiMsg m = (MidiMsg) e.nextElement();

			// Find out which row it should be in
			int r = 0;
			found = false;
			for ( Enumeration e2=notes.elements(); e2.hasMoreElements(); ) {
				MidiMsg m2 = (MidiMsg)e2.nextElement();
				if ( m2.getPitch() == m.getPitch() && m2.getChannel() == m.getChannel() ) {
					found = true;
					break;
				}
				r++;
			}
			if ( found ) {
				int c = (int)(m.getTime()/stepsize);
				setCell(r, c, cellLimit);
			}
			else {
System.out.println("Couldn't find row for m = "+m);
			}
		}
	}

	String getNameOf(int pitch, int chan, Vector drums ) {
		for ( Enumeration e=drums.elements(); e.hasMoreElements();){
			DrumInfo di = (DrumInfo)e.nextElement();
			MidiPhrase p = di.note;
			if ( p.size() != 1 )
				continue;
			MidiMsg m = p.peekFirstMsg();
			if ( m.getPitch() == pitch && m.getChannel() == chan )
				return di.label;
		}
		return null;
	}

	Vector getRows() {
		return rows;
	}

	int getNrows() {
		return rows.size();
	}

	public void setNsteps(int n) {
		setNsteps(n,true);
	}

	void setNsteps(int n, boolean fire) {
		int oldnsteps = nsteps;
		nsteps = n;
		makeEnoughCellsForSteps();
	}

	void changeDrum(int rownum, String drumName) {
		DrumRow dr = (DrumRow)rows.elementAt(rownum);
		boolean found = false;
		for ( Enumeration e=drums.elements(); e.hasMoreElements();){
			DrumInfo di = (DrumInfo)e.nextElement();
			if ( di.label.equals(drumName) ) {
				dr.label = drumName;
				dr.note = di.note;
				found = true;
				break;
			}
		}
		if ( found == false )
			System.out.println("changeDrum didn't find drumName="+drumName);
	}

	void clearRow(int r) {
		DrumRow dr = (DrumRow)rows.elementAt(r);
		for ( int c=0; c<dr.cells.length; c++ )
			changeCell(dr,r,c,0);
	}

	void setCell(int r, int c, int v) {
		DrumRow dr = (DrumRow)rows.elementAt(r);
		changeCell(dr,r,c,v);
	}

	public void toggleCell(int r, int c) {
		DrumRow dr = (DrumRow)rows.elementAt(r);
		changeCell(dr,r,c, (dr.cells[c] > 0 ? 0 : cellLimit) );
	}

	public void incrementCell(int r, int c) {
		DrumRow dr = (DrumRow)rows.elementAt(r);
		int n = dr.cells[c];
		if ( ++n >= (cellLimit+1) )
			n = 0;
		changeCell(dr,r,c,n);
	}

	private void changeCell(DrumRow dr, int r, int c, int v) {
		dr.cells[c] = v;
	}

	private void makeEnoughCellsForSteps() {
		int needed = nsteps;
		// add step as necessary to each row
		for ( int n=0; n<rows.size(); n++ ) {
			DrumRow dr = ((DrumRow)rows.elementAt(n));
			while ( dr.cells.length < needed ) {
				int[] newcells = new int[dr.cells.length+1];
				// copy values in old cells over to new one
				System.arraycopy(dr.cells,0,newcells,0,dr.cells.length);
				newcells[dr.cells.length] = 0;
				dr.cells = newcells;
			}
		}
	}

	public void setNrows(int needed) {
		setNrows(needed,true);
	}

	private void setNrows(int needed, boolean fire) {
		int oldrows = rows.size();
		while ( getNrows() < needed ) {
			DrumInfo di = (DrumInfo)(drums.elementAt(0));
			DrumRow dr = new DrumRow(di.label, di.note, nsteps);
			rows.addElement(dr);
		}
		while ( getNrows() > needed ) {
			rows.removeElementAt(getNrows()-1);
		}
	}

	public CellInfo getCell(int r, int c) {
		DrumRow dr = (DrumRow)rows.elementAt(r);
		return new CellInfo(r,c,dr);
	}

}
