/* NoSuch MIDI - Copyright 1997-1998 by Tim Thompson */

package com.nosuch.cooking;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.applet.*;
import com.nosuch.midi.MidiPhrase;
import com.nosuch.midi.MidiSystem;
import com.nosuch.midi.MidiException;
import com.nosuch.midi.MidiNoteOn;
import com.nosuch.midi.MidiNoteOff;
import com.nosuch.midi.MidiMsg;
import java.util.Random;
import java.lang.Math;

/**
 * A MIDI applet that presents a grid that makes music when you drag
 * the mouse over it.
 */

public class MouseMatrix extends Cooklet {

	int midiChan = 0;	// 0 to 15
	int midiPatch = 1;

	public String getName() {
		return "Mouse Matrix";
	}
	public Dimension getPreferredSize() {
		return new Dimension(400,300);
	}

	public Panel getPanel() {
		return new MouseMatrixPanel();
	}

	class MouseMatrixPanel extends Panel {
		MouseMatrixPanel() {
			// Add a border around the edge.
			setLayout(new BorderLayout(5,5));
			MouseMatrixPanel2 pp = new MouseMatrixPanel2();
			add("Center", pp);
			add("South", new Spacer());
			add("West", new Spacer());
			add("East", new Spacer());
			add("North", new Spacer());
			pp.reset();
		}
	}
	class Spacer extends Canvas {
		public Spacer() {
		}
		public Dimension getPreferredSize() {
			return new Dimension(0,0);
		}
	}
	class MouseMatrixPanel2 extends Panel {
		MouseMatrixControls controls;
		MouseMatrixPlayer player;
		MouseMatrixCanvas canvas;
		public MouseMatrixPanel2() {
			setLayout(new BorderLayout(5,5));
			RowCol rc = new RowCol(15,15);
			player = new MouseMatrixPlayer();
			canvas = new MouseMatrixCanvas(player,rc);
			controls = new MouseMatrixControls(player,canvas);
			add("South", controls);
			add("Center", canvas);
			add("West", new Spacer());
			add("East", new Spacer());
			add("North", new Spacer());
			player.setGridSize(rc);
		}
		public void reset() {
			controls.reset();
		}
	}
	public void writeExternal(ObjectOutput out) throws IOException {
	}
	public void readExternal(ObjectInput in) throws IOException {
	}
}

class MouseMatrixPlayer {

	public static final int STRUM = 0;
	public static final int STACCATO = 1;

	boolean ison = false;
	RowCol lastrc;
	int Mmmap[];
	int xinc1; int xinc2; int cols; int yinc1; int yinc2; int rows;
	int nnotes = 1;
	int mode = STRUM;
	MidiPhrase saveOff;

	public MouseMatrixPlayer() {
		saveOff = new MidiPhrase();
	}
	public void setNumNotes(int n) {
		nnotes = n;
	}
	public void setMode(int m) { mode = m; }
	public void setXinc1(int v) { xinc1 = v; mmmake(); }
	public void setXinc2(int v) { xinc2 = v; mmmake(); }
	public void setYinc1(int v) { yinc1 = v; mmmake(); }
	public void setYinc2(int v) { yinc2 = v; mmmake(); }

	public void setGridSize(RowCol rc) {
		cols = rc.col;
		rows = rc.row;
		mmmake();
	}
	void mmmake() {
		Mmmap = new int[(cols+1)*(rows+1)];
		int sp = xinc1;
		for ( int c=0; c<=cols; c++ ) {
			int thisp = sp;
			for ( int p=0; p<=rows; p++ ) {
				if ( thisp > 127 )
					thisp -= 128;
				Mmmap[c*rows+p] = thisp;
				if ( p%2 != 0 )
					thisp += yinc1;
				else
					thisp += yinc2;
			}
			if ( c%2 != 0 )
				sp += xinc1;
			else
				sp += xinc2;
		}
	}
	MidiPhrase mmmap(RowCol rc, int buttonnum, boolean noteon) {
		int t;
		if ( rc.row < 0 || rc.row > rows || rc.col < 0 || rc.col > cols ) {
			System.err.println("Bad row/col given to mmmap!");
			t = 0;
		}
		else
			t = rc.col*rows + rc.row;

		MidiPhrase p = new MidiPhrase();

		int p1 = Mmmap[t];
		int p2 = Mmmap[t+1];
		int p3 = Mmmap[t+2];
		int p4 = Mmmap[t+3];

		MidiMsg n1 = noteon ? (MidiMsg) new MidiNoteOn(p1) : (MidiMsg) new MidiNoteOff(p1);
		MidiMsg n2 = noteon ? (MidiMsg) new MidiNoteOn(p2) : (MidiMsg) new MidiNoteOff(p2);
		MidiMsg n3 = noteon ? (MidiMsg) new MidiNoteOn(p3) : (MidiMsg) new MidiNoteOff(p3);
		MidiMsg n4 = noteon ? (MidiMsg) new MidiNoteOn(p4) : (MidiMsg) new MidiNoteOff(p4);

		if ( nnotes == 4 ) {
			p.append(n1);
			p.append(n2);
			p.append(n3);
			p.append(n4);
		}
		else if ( nnotes == 3 ) {
			p.append(n1);
			p.append(n2);
			p.append(n3);
		}
		else if ( nnotes == 2 ) {
			p.append(n1);
			p.append(n2);
		}
		else
			p.append(n1);
		return p;
	}
	public synchronized void playcell(RowCol rc, boolean on) {

		try {
			if ( lastrc != null && rc.equals(lastrc) )  {
				// Same cell as last time - don't do anything if it's already on
				// and we get a request to turn it on.
				if ( ison == true && on == true )
					return;
			}
			else {
				// new cell - make sure previous cell is off in STACCATO mode
				if ( mode==STACCATO && ison == true ) {
					MidiSystem.schedule(saveOff);
					saveOff = new MidiPhrase();
				}
			}
	
			if ( on == true ) {
				MidiSystem.schedule(mmmap(rc,0,true));
				saveOff.append(mmmap(rc,0,false));
			}
			else {
				MidiSystem.schedule(saveOff);
				saveOff = new MidiPhrase();
			}
		} catch (MidiException e) {
			System.err.println("MidiException - "+e.getMessage());
		}
		lastrc = rc;
		ison = on;
	}
}
    
class MouseMatrixControls extends Panel implements ItemListener, ActionListener {

    MouseMatrixCanvas canvas;
	MouseMatrixPlayer player;
	Choice debugChoice;
	Choice noteChoice;
	Choice gridsizeChoice;
	Choice modeChoice;
	Choice xinc1Choice;
	Choice xinc2Choice;
	Choice yinc1Choice;
	Choice yinc2Choice;
	Button randButton;
	Button resetButton;

    public MouseMatrixControls(MouseMatrixPlayer p, MouseMatrixCanvas c) {
		canvas = c;
		player = p;

		setLayout(new GridLayout(0,5,5,5));

		add(randButton = new Button("Random"));
		randButton.addActionListener(this);
		xinc1Choice = makeChoice("Delta X1",1,12);
		xinc1Choice.addItemListener(this);
		xinc2Choice = makeChoice("Delta X2",1,12);
		xinc2Choice.addItemListener(this);

		add(resetButton = new Button("Reset"));
		resetButton.addActionListener(this);
		yinc1Choice = makeChoice("Delta Y1",1,12);
		yinc1Choice.addItemListener(this);
		yinc2Choice = makeChoice("Delta Y2",1,12);
		yinc2Choice.addItemListener(this);

		add(noteChoice = new Choice());
		noteChoice.addItem("1 note");
		noteChoice.addItem("2 notes");
		noteChoice.addItem("3 notes");
		noteChoice.addItem("4 notes");
		noteChoice.select(0);
		noteChoice.addItemListener(this);

		add(new Label("Grid Size"));
		add(gridsizeChoice = new Choice());
		for ( int i=10; i<=40; i+=5 ) {
			gridsizeChoice.addItem(Integer.toString(i));
		}
		gridsizeChoice.addItemListener(this);

		add(new Label("Mode"));
		add(modeChoice = new Choice());
		modeChoice.addItem("Strum");
		modeChoice.addItem("Staccato");
		modeChoice.addItemListener(this);
    }
	public void reset() {
		setXinc1(2);
		setXinc2(5);
		setYinc1(5);
		setYinc2(7);
		setGridSize(15);
	}
	public void setXinc1(int v) { player.setXinc1(v); xinc1Choice.select(v-1);}
	public void setXinc2(int v) { player.setXinc2(v); xinc2Choice.select(v-1);}
	public void setYinc1(int v) { player.setYinc1(v); yinc1Choice.select(v-1);}
	public void setYinc2(int v) { player.setYinc2(v); yinc2Choice.select(v-1);}
	public void setGridSize(int v) {
		String s = Integer.toString(v);
		gridsizeChoice.select(s);
		player.setGridSize(new RowCol(v,v));
		canvas.setGridSize(new RowCol(v,v));
	}

	public Choice makeChoice(String label, int valmin, int valmax){
		Choice c;
		add(new Label(label));
		add(c = new Choice());
		for ( int i=valmin; i<=valmax; i++ ) {
			c.addItem(Integer.toString(i));
		}
		return c;
	}

	public void itemStateChanged(ItemEvent e) {
		ItemSelectable s = e.getItemSelectable();
		String arg = (String)(e.getItem());
		if ( s == noteChoice ) {
			player.setNumNotes(noteChoice.getSelectedIndex()+1);
		}
		else if ( s == modeChoice ) {
			if ( modeChoice.getSelectedIndex() == 0 )
					player.setMode(MouseMatrixPlayer.STRUM);
			else
					player.setMode(MouseMatrixPlayer.STACCATO);
		}
		else if ( s == xinc1Choice )
			player.setXinc1(Integer.valueOf(arg).intValue());
		else if ( s == xinc2Choice )
			player.setXinc2(Integer.valueOf(arg).intValue());
		else if ( s == yinc1Choice )
			player.setYinc1(Integer.valueOf(arg).intValue());
		else if ( s == yinc2Choice )
			player.setYinc2(Integer.valueOf(arg).intValue());
		else if ( s == gridsizeChoice ) {
			setGridSize(Integer.valueOf(arg).intValue());
		}
	}

    public synchronized void actionPerformed(ActionEvent ev) {
		String s = ev.getActionCommand();
		if ( s.equals("Reset") ) {
			reset();
		}
		else if ( s.equals("Random") ) {
			Random r = new Random();
			setXinc1(Math.abs(r.nextInt()%12)+1);
			setXinc2(Math.abs(r.nextInt()%12)+1);
			setYinc1(Math.abs(r.nextInt()%12)+1);
			setYinc2(Math.abs(r.nextInt()%12)+1);
		}
    }
}

class RowCol {
	public int row;
	public int col;
	public RowCol(int r, int c) {
		row = r;
		col = c;
	}
	public boolean equals(RowCol rc) {
		if ( rc.row == row && rc.col == col )
			return true;
		else
			return false;
	}
}

class MouseMatrixCanvas extends Canvas implements MouseListener, MouseMotionListener {

	MouseMatrixPlayer player;
	int rows;
	int cols;

	public MouseMatrixCanvas(MouseMatrixPlayer p, RowCol rc) {
		player = p;
		rows = rc.row;
		cols = rc.col;
		setBackground(Color.lightGray);
		addMouseListener(this);
		addMouseMotionListener(this);
	}

	public void mouseClicked(MouseEvent e) { };
	public void mouseEntered(MouseEvent e) { };
	public void mouseExited(MouseEvent e) { };
	public void mousePressed(MouseEvent e) { };
	public void mouseReleased(MouseEvent e) { };
	public void mouseDragged(MouseEvent e) { };
	public void mouseMoved(MouseEvent e) { };

	public void setGridSize(RowCol rc) {
		rows = rc.row;
		cols = rc.col;
		repaint();
	}
	public void paint(Graphics g) {
		Rectangle r = getBounds();
		int ysz = r.height;
		int xsz = r.width;
		g.setColor(Color.black);
		g.drawRect(0,0,r.width-1,r.height-1);
		for ( int n=1; n<rows; n++ ) {
			int y = (n*ysz)/rows;
			g.drawLine(0,y,r.width,y);
		}
		for ( int n=1; n<cols; n++ ) {
			int x = (n*xsz)/cols;
			g.drawLine(x,0,x,r.height);
		}
	}
	public RowCol findrowcol(Point p) {
		Rectangle b = getBounds();
		int c = (p.x)*cols/(b.width);
		if ( c >= cols )
			c = cols-1;
		else if ( c < 0 )
			c = 0;
		int r = (p.y)*cols/(b.height);
		if ( r >= rows )
			r = rows-1;
		else if ( r < 0 )
			r = 0;
		return new RowCol(r,c);
	}

	public synchronized void processMouseMotionEvent(MouseEvent e) {
		if (e.getID() == MouseEvent.MOUSE_DRAGGED) {
			Point p = new Point(e.getX(),e.getY());
			RowCol rc = findrowcol(p);
			player.playcell(rc,true);
		}
	}

	public synchronized void processMouseEvent(MouseEvent e) {

		Point p = new Point(e.getX(),e.getY());
		RowCol rc;

		if (e.getID() == MouseEvent.MOUSE_PRESSED) {
			rc = findrowcol(p);
			player.playcell(rc,true);
		}
		else if (e.getID() == MouseEvent.MOUSE_RELEASED) {
			rc = findrowcol(p);
			player.playcell(rc,false);
		}
		super.processMouseEvent(e);
	}
}
