/* Copyright 1998 by Dexter Reid */

package com.nosuch.cooking;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import com.nosuch.midi.*;

/**
 * This class is just a thin wrapper to test the
 * NSPhraseView class.
 */
public class PianoRoll extends Cooklet implements MouseListener, MouseMotionListener, KeyListener {

	//
	// This is just for starters. The PianoRoll will
	//  have its own wrapper gui class for the phrase
	//  view, but to place development emphasis on NSPhraseView
	//  i'm leaving it out for now.
	//
	NSPhraseView m_phrase_view;
	Point m_press_pt;
	long m_scroll_rate_x;
	int m_scroll_rate_y;
	MidiPhrase m_phrase;


	public PianoRoll() {
		m_phrase_view = new NSPhraseView();
		m_phrase_view.addComponentListener(this);

		FileDialog fd = new FileDialog(new Frame(),"Load",FileDialog.LOAD);
		fd.setSize(400,200);
		fd.show();
		String fname = fd.getFile();
		String dir = fd.getDirectory();
		fname = dir + File.separator + fname;

		readMidiFromFile(fname);

		m_phrase_view.addMouseListener(this);
		m_phrase_view.addMouseMotionListener(this);
		m_phrase_view.addKeyListener(this);
	}

	protected void processComponentEvent(ComponentEvent e) {
	}

	public void readMidiFromFile(String file_path) {
		try {
			System.out.println("Reading "+file_path+" ...");
			m_phrase = MidiFileReader.readMidiPhrase(file_path);
			System.out.println("Done reading.");
			m_phrase_view.setPhrase(m_phrase, true);
		}
		catch (Exception ex) {
			ex.printStackTrace();
			System.out.println("Exception trying to read "+file_path+" msg="+ex.getMessage());
		}
	}

	public String getName() {
		return "NSPhraseView";
	}

	public Dimension getPreferredSize() {
		return new Dimension(800,300);
	}

	public Panel getPanel() {
		return m_phrase_view;
	}
	/**
	 * 
	 */
	public void mouseClicked(MouseEvent event) {
	}

	/**
	 * 
	 */
	public void mouseEntered(MouseEvent event) {
		//System.out.println("PianoRoll: mouse entered");
	}

	/**
	 * 
	 */

	public void mouseExited(MouseEvent event) {
	}

	/**
	 * 
	 */

	public void mousePressed(MouseEvent event) {
		//System.out.println(event);
		m_press_pt = event.getPoint();
		int [] temp_pitch = m_phrase_view.convertPitch(1);
		m_scroll_rate_y = temp_pitch[1] - temp_pitch[0];
		m_scroll_rate_y = m_scroll_rate_y > 1 ? m_scroll_rate_y:1;
		//m_scroll_rate_y = m_phrase_view.convertPitch(1);

	}

	/**
	 * 
	 */

	public void mouseReleased(MouseEvent event) {
	}

	/**
	 * 
	 */

	public void mouseDragged(MouseEvent event) {
		//m_phrase_view.panTimePitchRel(0,m_press_pt.y - event.getY());
		int [] temp_pitch = m_phrase_view.convertPitch(1);
		m_phrase_view.panXYRel(m_press_pt.x - event.getX(), m_scroll_rate_y * (event.getY() - m_press_pt.y) );
		m_press_pt = event.getPoint();
	}

	/**
	 * 
	 */

	public void mouseMoved(MouseEvent e) {
	}

	/**
	 * 
	 */

	public void keyTyped(KeyEvent e) {
		//System.out.println(e);
		switch(e.getKeyChar()) {
			case KeyEvent.VK_COMMA:
				m_phrase_view.panXYRel(-10,0);
				break;
			case KeyEvent.VK_PERIOD:
				m_phrase_view.panXYRel(10,0);
				break;
			case KeyEvent.VK_A:
				m_phrase_view.panTimePitchRel(0,2);
				break;
			case KeyEvent.VK_P:
				System.out.println("Scheduling the phrase...");
				try {
					MidiSystem.schedule(m_phrase);
				} catch (Exception ee) {
					System.out.println("Exception: msg="
						+ee.getMessage());
				}
				break;
			case KeyEvent.VK_Z:
				m_phrase_view.panTimePitchRel(0,-2);
				break;
			case KeyEvent.VK_CLOSE_BRACKET:
				m_phrase_view.zoomPercentXY(new Point(0,-1),110);
				break;
			case KeyEvent.VK_OPEN_BRACKET:
				m_phrase_view.zoomPercentXY(new Point(0,-1),90);
				break;
			case KeyEvent.VK_R:
				m_phrase_view.zoomToFitPhrase();
				break;
		}
	}

	/**
	 * 
	 */

	public void keyPressed(KeyEvent e) {
	}

	/**
	 * 
	 */

	public void keyReleased(KeyEvent e) {
	}
}
