/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

package com.nosuch.midi;

/**
 * Manage MIDI "text" messages, used to represent Standard MIDI File 
 * metatext and other miscellaneous things.
 *
 * @author Tim Thompson
 */

public class MidiText extends MidiMsg implements Cloneable {

	private String text;

	/**
	 * Create a MidiText message.
	 * @param t time (in clicks) of the message.
	 * @param s string of text in the message.
	 */
	public MidiText(long t, String s) {
		super(t);
		text = s;
	}

	/**
	 * Create a MidiText message.
	 * @param s string of text in the message.
	 */
	public MidiText(String s) {
		super(0);
		text = s;
	}

	public Object clone() {
		MidiText mb = new MidiText(time,text);
		return mb;
	}

	public boolean equals(MidiMsg m) {
		if ( ! super.equals(m) )
			return false;
		if ( ! ( m instanceof MidiText) )
			return false;
		MidiText mn = (MidiText)m;
		if ( ! mn.text.equals(text) )
			return false;
		return true;
	}

	/**
	 * Return the pitch of this message.
	 * For MidiByte messages, this is meaningless, so -1 is returned.
	 * This method (and the next few methods) are included so that
	 * code can be written conveniently using just the MidiMsg methods.
	 */
	public int getPitch() { return -1; }
	/**
	 * Set the pitch of this message.
	 * For MidiByte messages, this is meaningless, so it is ignored.
	 */
	public void setPitch(int p) { }

	/**
	 * Return the duration of this message.
	 * For MidiByte messages, this is meaningless, so -1 is returned.
	 */
	public long getDuration() { return 0; }	// keep this 0, for append()

	/**
	 * Set the duration of this message.
	 * For MidiByte messages, this is meaningless, so it is ignored.
	 */
	public void setDuration(long d) { }

	/**
	 * Return the channel of this message.
	 * For MidiText messages, this is is potentially useful (e.g. for
	 * controller messages and things), but for the moment we return -1.
	 */
	public int getChannel() { return -1; }

	/**
	 * Set the channel of this message.
	 * For MidiText messages, this is is potentially useful (e.g. for
	 * controller messages and things), but for the moment we ignore it.
	 */
	public void setChannel(int c) { }

	/**
	 * Return the velocity of this message.
	 * For MidiText messages, this is meaningless, so we return -1.
	 */
	public int getVelocity() { return -1; }

	/**
	 * Set the velocity of this message.
	 * For MidiText messages, this is meaningless, so we ignore it.
	 */
	public void setVelocity(int v) { }

	/**
	 * Return an array of bytes containing the contents of the message.
	 * This is used when generating actual MIDI output, so for "text"
	 * messages this is null.
	 */
	public byte[] getBytesOn() { return null; }

	/**
	 * Return an array of bytes containing the ending half of the message.
	 */
	public byte[] getBytesOff() { return null; }

	/**
	 * Return a human-readable representation of the message.
	 */
	public String toString() {
		String s = "MidiText(t="+Long.toString(time)+",text="+text+")";
		return s;
	}
}
