/* NoSuch MIDI - Copyright 1997 by Tim Thompson */

/*
 * Windows MIDI methods, JNI style.
 */

#include <stdio.h>
#include <windows.h>
/*
#include <afxwin.h>
#include <afxext.h>
#include <mmsystem.h>
*/

#include <jni.h>
#include <oobj.h>
#include <interpreter.h>

#include "com_nosuch_midi_MidiSystem.h"
#include "com_nosuch_midi_Win32MidiOutput.h"
#include "com_nosuch_midi_Win32MidiInput.h"

#define WM_MIDI_TIMEOUT   (WM_USER + 11)
#define WM_MIDI_ERROR    (WM_USER + 21)
#define WM_MIDIINPUT    (WM_USER + 31)
#define WM_MIDIOUTPUT    (WM_USER + 41)

typedef struct event_tag {
	DWORD device;
	DWORD timestamp;
	DWORD data;	/* for short events, this is the raw data */
			/* for long events, this is a pointer to MIDIHDR */
	WORD islong;
	UINT msg;	/* WM_* message type */
} EVENT;
typedef EVENT FAR *LPEVENT;

/* Structure to manage the circular input buffer.  */

typedef struct BufferList_tag {
	LPEVENT list;
	LPEVENT endoflist;
	LPEVENT head;
	LPEVENT tail;
	DWORD   total;
	DWORD   sofar;
} BufferList;

typedef BufferList FAR *LPBufferList;

/* Structure to pass instance data from the application
 * to the low-level callback function.
 */
typedef struct callbackInstance_tag
{
    DWORD               device;
    LPBufferList    bufflist;
} MidiCallback;

typedef MidiCallback FAR *LPMidiCallback;

#ifdef __cplusplus
extern "C" {
#endif
void FAR PASCAL midiInputHandler(HMIDIIN, WORD, DWORD, DWORD, DWORD);
void FAR PASCAL midiOutputHandler(HMIDIIN, WORD, DWORD, DWORD, DWORD);
extern HANDLE midiInEvent;
#ifdef __cplusplus
}
#endif

void FAR PASCAL MidiPutEvent(LPBufferList bufflist, LPEVENT lpEvent);
LPMidiCallback FAR PASCAL AllocCallbackInstanceData(void);
void FAR PASCAL FreeCallbackInstanceData(LPMidiCallback bufflist);
LPBufferList AllocCircularBuffer(DWORD total);
void FreeCircularBuffer(LPBufferList bufflist);
WORD FAR PASCAL MidiGetEvent(LPBufferList bufflist, LPEVENT lpEvent);

#ifdef __cplusplus
extern "C" {
#endif

extern int DebugMidi;

BOOL initmidi(JNIEnv *);
int init_midiout(JNIEnv *);
int init_midiin(JNIEnv *);
typedef unsigned char BYTE;
char *writemidi(long,int,BYTE*);
int readmidi(BYTE*,int);
HMIDIOUT openmidiout(JNIEnv*,int);
HMIDIIN openmidiin(JNIEnv*,int);
void throwMidiException(JNIEnv*,char *);

HANDLE midiInEvent;
HANDLE midiOutEvent;

#ifdef __cplusplus
}
#endif

BOOL native_initmidi(JNIEnv *env);

class ScheduledEvent {
public:
	long time;
	BYTE *bytes;
	int leng;
	class ScheduledEvent *next;
};
ScheduledEvent* TopSched = 0;

// Minimum tempo value is 10 beats per second - this is perhaps a tad slower
// than it has to be, but things get out of hand quickly if it goes faster.
#define MINTEMPO 100000

long Tempo = 500000;// in microseconds.	Is # of microseconds per beat
long MillTempo = 500;// ie. Tempo/1000
long Clicks = 96;	// in clicks.		Is # of clicks per beat
long Now;			// in clicks.		Is current time
jlong StartMilli;	// in milliseconds - beginning of application
			// This value is actually adjusted when the tempo is
			// changed, so that Now can be computed from it.

void
throwMidiException(JNIEnv *env, char *msg)
{
	jclass cls = env->FindClass("midi/MidiException");
	if ( cls != 0 )
		env->ThrowNew(cls, msg);
}

void
throwNullException(JNIEnv *env, char *msg)
{
	jclass cls = env->FindClass("java/lang/NullPointerException");
	if ( cls != 0 )
		env->ThrowNew( cls, msg);
}

/***********
 * Win32MidiOutput native methods
 **********/
JNIEXPORT long JNICALL
Java_com_nosuch_midi_MidiSystem_nativeMidiOutGetNumDevs(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )
		return -1;
	int n = midiOutGetNumDevs(); 
	return n;
}

jint
getDeviceID(JNIEnv *env,jobject obj)
{
	jclass cls = env->GetObjectClass(obj);
	if ( cls == 0 )
		return -1;
	jfieldID fid_deviceId = env->GetFieldID(cls,"deviceId","I");
// printf("getDeviceID, jid_deviceId = %ld\n",(long)fid_deviceId);
	if ( fid_deviceId == 0 )
		return -1;
	jint ji = env->GetIntField(obj,fid_deviceId);
	return ji;
}

JNIEXPORT void JNICALL
Java_com_nosuch_midi_Win32MidiOutput_MidiOutOpen(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )	// throws exception
		return;

	jint devid = getDeviceID(env,obj);
	// It is NOT an error if devid is -1 - that's MIDIMAPPER
	int nout = midiOutGetNumDevs();
	if ( (devid < -1) || (devid >= nout) ) {
		throwMidiException(env,"Bad deviceId in MidiOutOpen");
		return;
	}

	HMIDIOUT h = openmidiout(env,(int)devid);	// throws exception
	if ( h == NULL )
		return;

	jclass cls = env->GetObjectClass(obj);
	if ( cls == 0 ) {
		throwMidiException(env,"Can't get class in MidiOutOpen");
		return;
	}
	jfieldID fid_deviceHandle = env->GetFieldID(cls,"deviceHandle","I");
	if ( fid_deviceHandle == 0 ) {
		throwMidiException(env,"Can't find deviceHandle field in MidiOutOpen");
		return;
	}
	env->SetIntField(obj,fid_deviceHandle,(jint)h);
	return;
}

jint
getDeviceHandle(JNIEnv *env, jobject obj)
{
	jclass cls = env->GetObjectClass(obj);
	if ( cls == 0 ) {
		throwMidiException(env,"can't get class!");
		return 0;
	}
	jfieldID fid_deviceHandle = env->GetFieldID(cls,"deviceHandle","I");
	if ( fid_deviceHandle == 0 ) {
		throwMidiException(env,"can't get deviceHandle field");
		return 0;
	}
	jint ji = env->GetIntField(obj,fid_deviceHandle);
	return ji;
}

JNIEXPORT long JNICALL
Java_com_nosuch_midi_Win32MidiOutput_MidiOutReset(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )
		return -1;
	jint h = getDeviceHandle(env,obj);
	if ( h == 0 )
		return -1;
	long l = midiOutReset((HMIDIOUT)h);
	if ( l != MMSYSERR_NOERROR ) {
		throwMidiException(env,"MidiOutReset returns error");
		return -1;
	}
	return 0;
}

JNIEXPORT long JNICALL
Java_com_nosuch_midi_Win32MidiOutput_MidiOutClose(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )
		return -1;
	jint h = getDeviceHandle(env,obj);
	if ( h == 0 )
		return -1;
	long l = midiOutClose((HMIDIOUT)h);
	if ( l != MMSYSERR_NOERROR ) {
		throwMidiException(env,"MidiOutClose returns error");
		return -1;
	}
	return 0;
}


/***********
 * Win32MidiInput native methods
 **********/

JNIEXPORT long JNICALL
Java_com_nosuch_midi_MidiSystem_nativeMidiInGetNumDevs(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )
		return -1;
	int n = midiInGetNumDevs(); 
	return n;
}

JNIEXPORT long JNICALL
Java_com_nosuch_midi_Win32MidiInput_MidiInOpen(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )
		return -1;

	jint devid = getDeviceID(env,obj);
	if ( devid < 0 )
		return -1;

	HMIDIIN h = openmidiin(env, (int)devid);
	if ( h == NULL )
		return -1;

	jclass cls = env->GetObjectClass(obj);
	if ( cls == 0 ) {
		throwMidiException(env,"can't get class!");
		return -1;
	}
	jfieldID fid_deviceHandle = env->GetFieldID(cls,"deviceHandle","I");
	if ( fid_deviceHandle == 0 )
		return -1;
	env->SetIntField(obj,fid_deviceHandle,(jint)h);
	return 0;
}

JNIEXPORT long JNICALL
Java_com_nosuch_midi_Win32MidiInput_MidiInReset(JNIEnv *env, jobject obj)
{
	if ( native_initmidi(env) == FALSE )
		return -1;
	jint h = getDeviceHandle(env,obj);
	if ( h == 0 )
		return -1;
	MMRESULT r = midiInReset((HMIDIIN)h);
	if ( r != MMSYSERR_NOERROR ) {
		throwMidiException(env,"MidiInReset returns error");
		return -1;
	}
	return 0;
}

JNIEXPORT long JNICALL
Java_com_nosuch_midi_Win32MidiInput_MidiInClose(JNIEnv *env, jobject obj)
{

	if ( native_initmidi(env) == FALSE )
		return -1;

	jint h = getDeviceHandle(env,obj);
	if ( h == 0 )
		return -1;
	MMRESULT r = midiInClose((HMIDIIN)h);
	if ( r != MMSYSERR_NOERROR ) {
		throwMidiException(env,"MidiInClose returns error");
		return -1;
	}
	return 0;
}

static BOOL midiInputRunning = FALSE;

JNIEXPORT void JNICALL
Java_com_nosuch_midi_Win32MidiInput_MidiInLoopStop(JNIEnv *env, jobject obj)
{
	midiInputRunning = FALSE;
	::SetEvent(midiInEvent);
}

static BOOL
makeHighPriorityThread()
{
	BOOL pr1 = SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS);
	BOOL pr2 = SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_TIME_CRITICAL);
	if ( pr1 == FALSE || pr2 == FALSE )
		return FALSE;
	return TRUE;
}

JNIEXPORT void JNICALL
Java_com_nosuch_midi_Win32MidiInput_MidiInLoop(JNIEnv *env, jobject obj, jboolean makerealtime)
{
	if ( native_initmidi(env) == FALSE )
		return;

	if ( midiInputRunning == TRUE ) {
		throwMidiException(env,"Multiple MidiInLoops not allowed");
		return;
	}

	midiInputRunning = TRUE;

	midiInEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);

	jclass cls = env->GetObjectClass(obj);
	if ( cls == 0 ) {
		throwMidiException(env,"Can't find class");
		return;
	}
	jmethodID meth = env->GetMethodID(cls,"gotByte","(I)V");
	if ( meth == 0 ) {
		throwMidiException(env,"Can't find gotByte method");
		return;
	}

	if ( makerealtime == TRUE ) {
		if ( makeHighPriorityThread() == FALSE ) {
			throwMidiException(env,"can't set thread to high priority");
			return;
		}
	}

	while (TRUE) {

		::WaitForSingleObject(midiInEvent, 1000);

	    	if ( midiInputRunning == FALSE )
			break;

		BYTE buff[100];
		int nb;
		while ( (nb=readmidi(buff,100)) != 0 ) {
			for ( int n=0; n<nb; n++ ) {
				jint bb = buff[n];
				env->CallVoidMethod(obj, meth, bb);
			}
		}
	}
}

JNIEXPORT void JNICALL
Java_com_nosuch_midi_MidiSystem_nativeSetDebug(JNIEnv *env, jobject obj, jint t)
{
	DebugMidi = t;
}

JNIEXPORT void JNICALL
Java_com_nosuch_midi_MidiSystem_nativeSetTempo(JNIEnv *env, jobject obj, jint tempo)
{
	if ( tempo <= 0 ) {
		throwMidiException(env,"invalid tempo");
		return;
	}

	// This is in preparation for adjusting StartMilli
	long tm = timeGetTime();
	long oldNow = Clicks * (long)(tm-StartMilli) / MillTempo;

	Tempo = tempo;

	if ( Tempo < MINTEMPO )
		Tempo = MINTEMPO;
	MillTempo = Tempo/1000;

	// Adjust StartMilli so it can still be used to compute Now
	StartMilli = tm - ((oldNow * MillTempo)/Clicks);
}


JNIEXPORT void JNICALL
Java_com_nosuch_midi_MidiSystem_nativeSetClicks(JNIEnv *env, jobject obj, jint c)
{
	if ( c <= 0 ) {
		throwMidiException(env,"invalid clicks");
		return;
	}
	Clicks = c;
}

static BOOL midiOutputRunning = FALSE;

JNIEXPORT void JNICALL
Java_com_nosuch_midi_Win32MidiOutput_MidiOutLoopStop(JNIEnv *env, jobject obj)
{
	midiOutputRunning = FALSE;
	::SetEvent(midiOutEvent);
}

JNIEXPORT void JNICALL
Java_com_nosuch_midi_Win32MidiOutput_MidiOutLoop(JNIEnv *env, jobject obj, jboolean makerealtime)
{
	if ( native_initmidi(env) == FALSE )
		return;

	if ( midiOutputRunning == TRUE ) {
		throwMidiException(env,"Multiple MidiOutLoops not allowed");
		return;
	}

	midiOutputRunning = TRUE;

	midiOutEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);

	// Get deviceHandle value of midiOut field.

	jclass cls;
	jfieldID fid;
	jint hand;

	if ( (cls=env->GetObjectClass(obj)) == 0
		|| (fid=env->GetFieldID(cls,"deviceHandle","I")) == 0
		|| (hand=env->GetIntField(obj,fid)) == 0 ) {
		throwMidiException(env,"can't get deviceHandle field");
		return;
	}
	long tmout;

	StartMilli = timeGetTime();
	Now = 0;
	if ( makerealtime == TRUE ) {
		if ( makeHighPriorityThread() == FALSE ) {
			throwMidiException(env,"can't set thread to high priority");
			return;
		}
	}
	
	while ( TRUE ) {

		// Compute Now, the current time in clicks.

		// This can be optimized by computing the millitime() of the
		// next click, and just comparing the times.

		if ( TopSched != 0 ) {
			tmout = (TopSched->time - Now)*((MillTempo)/Clicks);
			if ( tmout < 0 )
				tmout = 0;
		}
		else
			tmout = 5000;

		::WaitForSingleObject(midiOutEvent, tmout);
		if ( midiOutputRunning == FALSE )
			break;

		Now = Clicks * (long)(timeGetTime()-StartMilli) / MillTempo;

		if ( TopSched == 0 )
			continue;

		if ( TopSched->time <= Now ) {
			ScheduledEvent *e = TopSched;
			TopSched = e->next;
			char *err = writemidi(hand,e->leng,e->bytes);
			if ( err != 0 )
				throwMidiException(env,err);
			delete [] e->bytes;
			delete e;
		}
	}
}

JNIEXPORT void JNICALL
Java_com_nosuch_midi_Win32MidiOutput_scheduleBytes(JNIEnv *env, jobject obj, jlong tm, jbyteArray bytearr) {

	ScheduledEvent *ne;
	jsize thelength;
	static int inside = 0;

	if ( native_initmidi(env) == FALSE )
		return;

	if ( inside != 0 )
		return;

	inside = 1;

	if ( bytearr == 0 ) {
		throwNullException(env,"null byte array");
		inside = 0;
		return;
	}
	if ( tm < 0 ) {
		throwMidiException(env,"negative time");
		inside = 0;
		return;
	}

	thelength = env->GetArrayLength(bytearr);

	ne = new ScheduledEvent();
	ne->time = (long)tm;
	ne->bytes = new BYTE[thelength];	// save a copy of the bytes
	ne->leng = thelength;

	// copy from java space to our copy
	env->GetByteArrayRegion(bytearr,0,thelength,(jbyte*)(ne->bytes));

	// Insert into schedule list
	if ( TopSched == 0 || tm < TopSched->time ) {
		ne->next = TopSched;
		TopSched = ne;
	}
	else {
		ScheduledEvent *prev = TopSched;
		ScheduledEvent *s = TopSched->next;
		for ( ; s != 0; prev=s,s=s->next ) {
			if ( tm < s->time )
				break;
		}
		// Insert ne after prev
		ne->next = prev->next;
		prev->next = ne;
	}

// printf("schedule SetEvent(midiOut)\n");
	::SetEvent(midiOutEvent);
// printf("after SetEvent(midiOut)\n");

	inside = 0;
}

static BOOL Midiinit = FALSE;

BOOL
native_initmidi(JNIEnv *env)
{
	// This stuff should only be done once, NOT every time an applet
	// starts up.

	if ( Midiinit == TRUE )
		return TRUE;

	// Watch out, the order here is needed to make sure Buffer_midiin
	// is initialized, because it's re-used (temporarily) in initmidi_output().
	if ( init_midiin(env) != 0 )
		return FALSE;
	if ( init_midiout(env) != 0 )
		return FALSE;
	Midiinit = TRUE;
	return TRUE;
}

JNIEXPORT jlong JNICALL
Java_com_nosuch_midi_MidiSystem_nativeNow(JNIEnv *env, jobject obj)
{
	Now = Clicks * (long)(timeGetTime()-StartMilli) / MillTempo;
	return Now;
}

#define NOTEOFF 0x80
#define NOTEON 0x90
#define PRESSURE 0xa0
#define CONTROLLER 0xb0
#define PITCHBEND 0xe0
#define PROGRAM 0xc0
#define CHANPRESSURE 0xd0

#define PORTAMENTO 0x41

#define SYSTEMMESSAGE	0xf0
#define BEGINSYSEX	0xf0
#define MTCQUARTERFRAME	0xf1
#define SONGPOSPTR	0xf2
#define SONGSELECT	0xf3

#define CIRCULAR_BUFF_SIZE 1000
#define NUM_OF_OUT_BUFFS 4
#define MIDI_OUT_BUFFSIZE 32
#define NUM_OF_IN_BUFFS 4
#define MIDI_IN_BUFFSIZE 8096
#define EBUFFSIZE 128
#define MAX_MIDI_DEVICES 8

int Moutinit = 0;
int Mininit = 0;
int DebugMidi = 0;
long Lasttimeout;

HINSTANCE MidiInstance = 0;
HMIDIIN Handle_midiin;
HMIDIOUT Handle_midiout;
MIDIINCAPS midiInCaps[MAX_MIDI_DEVICES];
MIDIOUTCAPS midiOutCaps[MAX_MIDI_DEVICES];
int Nmidiin = 0;
int Nmidiout = 0;
LPBufferList Buffer_midiin;

typedef struct mymidiinbuff {
	LPMIDIHDR midihdr;
} MY_MIDI_BUFF;

static MY_MIDI_BUFF *MidiInBuff = NULL;

LPMidiCallback CallbackInputData = NULL;
LPMidiCallback CallbackOutputData = NULL;

void handlemidioutput(long);
static LPMIDIHDR newmidihdr(int);
static int InitMidiInBuff();
static void freemidihdr(LPMIDIHDR);
static void midiouterror(JNIEnv *env, int r, char *s);
static void midiinerror(JNIEnv *env, int r, char *s);
int init_midiin(JNIEnv *env);
int init_midiout(JNIEnv *env);

static unsigned char *bytes = NULL;
static int sofar;
static int nbytes;
static int midiTimerId = 0;

int
readmidi(BYTE *buff,int buffsize)
{
	static EVENT ev;
	static unsigned char smallbuff[3];
	static int waslong;
	LPMIDIHDR lpMidi;
	int r;

	if ( ! Mininit )
		return 0;
	if ( bytes == NULL ) {
		if ( MidiGetEvent(Buffer_midiin,&ev) == 0 )
			return 0;
		if ( ev.msg != WM_MIDIINPUT ) {
			printf("Hey, ev.msg (%d) != WM_MIDIINPUT in getnmidi!?",ev.msg);
			return 0;
		}
		else if ( ev.islong ) {
			lpMidi = (LPMIDIHDR)ev.data;
			if ( ((lpMidi->dwFlags) & MHDR_DONE) == 0 ) {
				printf("Hey, MHDR_DONE not set in dwFlags!?");
			}
			bytes = (unsigned char*)(lpMidi->lpData);
			nbytes = (int)(lpMidi->dwBytesRecorded);
			sofar = 0;
		}
		else {
			BYTE statusbyte;
			smallbuff[0] = LOBYTE(LOWORD(ev.data));
			statusbyte = smallbuff[0] & (BYTE)0xf0;
			switch ( statusbyte ) {

			/* Three byte events */
			case NOTEOFF:
			case NOTEON:
			case PRESSURE:
			case CONTROLLER:
			case PITCHBEND:
				nbytes = 3;
				break;

			/* Two byte events */
			case PROGRAM:
			case CHANPRESSURE:
				nbytes = 2;
				break;

			/* MIDI system events (0xf0 - 0xff) */
			case SYSTEMMESSAGE:
			    switch(statusbyte) {
				/* Two byte system events */
				case MTCQUARTERFRAME:
				case SONGSELECT:
					nbytes = 2;
					break;
				/* Three byte system events */
				case SONGPOSPTR:
					nbytes = 3;
					break;
				/* One byte system events */
				default:
					nbytes = 1;
					break;
			    }
			    break;
            
			default:
				/* unknown event !? */
				nbytes = 0;
				break;
			}
			if ( nbytes > 1 ) {
				smallbuff[1] = HIBYTE(LOWORD(ev.data));
				if ( nbytes > 2 )
					smallbuff[2] = LOBYTE(HIWORD(ev.data));
			}
			bytes = smallbuff;
			sofar = 0;
		}
	}
	r = 0;
// printf("getnmidi mid\n");
	if ( bytes != NULL ) {
		BYTE *bp = buff;
		while ( r < buffsize && sofar < nbytes ) {
			*bp++ = bytes[sofar++];
			r++;
		}
		if ( sofar >= nbytes ) {
			/* We've used up all the bytes */
			bytes = NULL;
			if ( ev.islong ) {
				/* give buffer back to the MIDI input device */
				int i = (int)(ev.device);
				if ( i < 0 || i >= Nmidiin ) {
					printf("Hey, bad ev.device!?");
				}
				else {
					midiInPrepareHeader(Handle_midiin,
						(LPMIDIHDR)(ev.data),sizeof(MIDIHDR));
					midiInAddBuffer(Handle_midiin,
						(LPMIDIHDR)(ev.data),sizeof(MIDIHDR));
				}
			}
		}
	}
// printf("getnmidi end r=%d\n",r);
	return r;
}

char *
writemidi(long deviceHandle,int n,BYTE *p)
{
	static int inside = 0;
	BYTE *q;
	LPMIDIHDR midihdr;
	int r;
	union {
		DWORD dwData;
		BYTE bData[4];
	} u;

// printf("putnmidi n=%d\n",n);
	if ( inside != 0 ) {
		return "Recursive putnmidi ignored";
	}
	inside = 1;

	HMIDIOUT h = (HMIDIOUT)deviceHandle;

	if ( ! Moutinit ) {
		inside = 0;
		return "Moutinit not set in putnmidi";
	}
	if ( 0 && n <= 3 ) {
		u.bData[0] = (n>0) ? p[0] : 0;
		u.bData[1] = (n>1) ? p[1] : 0;
		u.bData[2] = (n>2) ? p[2] : 0;
		u.bData[3] = 0;
		r = midiOutShortMsg(h, u.dwData);
		if ( r ) {
			inside = 0;
			return "Error in midiOutShortMsg";
		}
	}
	else {
// printf("putnmidi long n=%d\n",n);
		midihdr = newmidihdr(n);
		if ( midihdr == 0 ) {
			inside = 0;
			return "Unable to allocate midihdr in putnmidi";
		}
		r = midiOutPrepareHeader(h, midihdr,sizeof(MIDIHDR));
		if ( r ) {
			inside = 0;
			return "Error in midiOutPrepareHeader";
		}
		for ( q=(BYTE*)(midihdr->lpData); n-- > 0; ) {
			*q++ = *p++;
		}
		r = midiOutLongMsg(h,midihdr,sizeof(MIDIHDR));
		if ( r ) {
			inside = 0;
			return "Error in midiOutLongMsg";
		}
	}
	inside = 0;
	return 0;
}

HMIDIIN
openmidiin(JNIEnv *env, int i)
{
	int r;

// printf("IN midi_misc.cpp, openmidiin, i=%d\n",i);
	r = midiInOpen((LPHMIDIIN)&Handle_midiin,
			  i,
			  (DWORD)midiInputHandler,
			  (DWORD)CallbackInputData,
			  CALLBACK_FUNCTION);
	if(r) {
		/* can't open */
		FreeCallbackInstanceData(CallbackInputData);
		midiinerror(env, r,"Error in midiInOpen: ");
		return 0;
	}
	else {
		if ( InitMidiInBuff() ) {
			midiInStart(Handle_midiin);
			Mininit = 1;
		}
		else {
			throwMidiException(env,"Unable to allocate Midi Buffers!?");
			return 0;
		}
	}
	return Handle_midiin;
}

HMIDIOUT
openmidiout(JNIEnv *env, int devno)
{
	int r;

	if ( devno >= Nmidiout ) {
		midiouterror(env,-1,"Bad output device number");
		return 0;
	}

	/* The magical number -1 is MIDIMAPPER */
	if ( devno == -1 )
		devno = MIDIMAPPER;

	/* Only one output open at a time */
	if ( Moutinit ) {
		midiOutClose(Handle_midiout);
		Moutinit = 0;
	}

	r = midiOutOpen ( &Handle_midiout, devno,
		(DWORD)midiOutputHandler,
		(DWORD)CallbackOutputData,
		(DWORD) CALLBACK_FUNCTION);

	if ( r ) {
		midiouterror(env,r,"Error in midiOutOpen: ");
		Handle_midiout = 0;	/* to make sure */
		Moutinit = 0;
	}
	else
		Moutinit = 1;
	return Handle_midiout;
}

static void
closemidiin()
{
	midiInReset(Handle_midiin);
	midiInClose(Handle_midiin);
	Mininit = 0;
}

UINT Millires = 5;

void
closemidi()
{
	if ( Moutinit ) {
		timeEndPeriod(Millires);
		midiOutClose(Handle_midiout);
		closemidiin();
	}
}

int
init_midiin(JNIEnv *env)
{
	int r, i;

	/* Allocate a circular buffer for low-level MIDI input.  This buffer
	 * is filled by the low-level callback function and emptied by the
	 * application when it receives MM_MIDIINPUT messages.
	 */
	Buffer_midiin = AllocCircularBuffer((DWORD)(CIRCULAR_BUFF_SIZE));
	if (Buffer_midiin == NULL) {
		throwMidiException(env,"Not enough memory available for input buffer.");
		return 1;
	}

	/* Get the number of MIDI input devices.  Then get the capabilities of
	 * each device.
	 */
	Nmidiin = midiInGetNumDevs();
	if (!Nmidiin) {
		throwMidiException(env,"There are no MIDI input devices.");
		return 1;
	}
	for (i=0; (i<Nmidiin) && (i<MAX_MIDI_DEVICES); i++) {
		r = midiInGetDevCaps(i, (LPMIDIINCAPS) &midiInCaps[i],
                                sizeof(MIDIINCAPS));
		if(r) {
			midiinerror(env, r,"Error in midiInGetDevCaps: ");
		}
	}

	if ( i > 0 ) {
		
		if ((CallbackInputData = AllocCallbackInstanceData()) == NULL) {
			printf("Not enough memory available.");
			FreeCircularBuffer(Buffer_midiin);
			return 1;
		}
		CallbackInputData->device = 0;
		CallbackInputData->bufflist = Buffer_midiin;
	}
	return 0;
}

int
init_midiout(JNIEnv *env)
{
	int r, i;

	if ( CallbackOutputData == NULL ) {
		if ((CallbackOutputData = AllocCallbackInstanceData()) == NULL) {
			throwMidiException(env,"Not enough memory available!?");
			return 1;
		}
		CallbackOutputData->device = 0;
		// For the moment, we use the Buffer_midiin event list for
		// sending back output buffer release events (WM_MIDIOUTPUT).
		CallbackOutputData->bufflist = Buffer_midiin;
	}

	Nmidiout = midiOutGetNumDevs();
	for (i=0; (i<Nmidiout) && (i<MAX_MIDI_DEVICES); i++) {
		r = midiOutGetDevCaps(i, (LPMIDIOUTCAPS) &midiOutCaps[i],
                                sizeof(MIDIOUTCAPS));
		if(r) {
			midiouterror(env,r,"Error in midiOutGetDevCaps: ");
			Nmidiout = i;		/* and admit no more devices */
			break;
		}
	}
	if ( Nmidiout == 0 ) {
		printf("No MIDI devices found - don't expect any output!");
		return 1;
	}
	return 0;
}

static LPMIDIHDR
newmidihdr(int sz)
{
	LPMIDIHDR lpMidi;
	HPSTR lpData;

	lpData = (HPSTR) GlobalAlloc(0, (DWORD)sz);
	if(lpData == NULL)
		return 0;
	    
	lpMidi = (LPMIDIHDR) GlobalAlloc(0, (DWORD)sizeof(MIDIHDR));
	if(lpMidi == NULL)
		return 0;
	    
	lpMidi->lpData = lpData;
	lpMidi->dwBufferLength = sz;
	lpMidi->dwBytesRecorded = 0;
	lpMidi->dwUser = 0;
	lpMidi->dwFlags = 0;
	return lpMidi;
}

static void
freemidihdr(LPMIDIHDR midihdr)
{
	GlobalFree(midihdr->lpData);
	GlobalFree(midihdr);
}

void
handlemidioutput(long lParam)
{
	LPMIDIHDR midihdr;
	if ( Moutinit ) {
		midihdr = (LPMIDIHDR) lParam;
		midiOutUnprepareHeader(Handle_midiout, midihdr, sizeof(MIDIHDR));
		freemidihdr(midihdr);
	}
}

static int
InitMidiInBuff()
{
	int n;

	if ( MidiInBuff )
		return 1;

	MidiInBuff = (MY_MIDI_BUFF*)malloc( NUM_OF_IN_BUFFS * sizeof (MY_MIDI_BUFF) );
	if ( MidiInBuff == NULL )
		return 0;

	for ( n=0; n<NUM_OF_IN_BUFFS; n++ ) {
	
		MidiInBuff[n].midihdr = newmidihdr(MIDI_IN_BUFFSIZE);

		midiInPrepareHeader(Handle_midiin,
			MidiInBuff[n].midihdr,sizeof(MIDIHDR));
		midiInAddBuffer(Handle_midiin,
			MidiInBuff[n].midihdr,sizeof(MIDIHDR));
	}
	return 1;
}

/*
 * circbuf.c - Routines to manage the circular MIDI input buffer.
 *      This buffer is filled by the low-level callback function and
 *      emptied by the application.  Since this buffer is accessed
 *      by a low-level callback, memory for it must be allocated
 *      exactly as shown in AllocCircularBuffer().
 */

/*
 * AllocCircularBuffer -    Allocates memory for a BufferList structure 
 * and a buffer of the specified size.
 *
 * Params:  total - The size of the buffer, in events.
 *
 * Return:  A pointer to a BufferList structure identifying the 
 *      allocated display buffer.  NULL if the buffer could not be allocated.
 */

static LPBufferList
AllocCircularBuffer(DWORD total)
{
    LPBufferList bufflist;
    LPEVENT lpMem;
    
    bufflist = (LPBufferList) GlobalAlloc(0,(DWORD)sizeof(BufferList));
    if ( bufflist == NULL )
        return NULL;
    
    lpMem = (LPEVENT) GlobalAlloc(0, total * sizeof(EVENT));
    if ( lpMem == NULL )
        return NULL;
    
    bufflist->list = lpMem;
    bufflist->endoflist = lpMem + total;
    bufflist->total = total;
    bufflist->sofar = 0L;
    bufflist->tail = lpMem;
    bufflist->head = lpMem;
        
    return bufflist;
}

/* FreeCircularBuffer - Frees the memory for the given BufferList 
 * structure and the memory for the buffer it references.
 *
 * Params:  bufflist - Points to the BufferList to be freed.
 *
 * Return:  void
 */

static void
FreeCircularBuffer(LPBufferList bufflist)
{
    GlobalFree(bufflist->list);
    GlobalFree(bufflist);
}

/* MidiGetEvent - Gets a MIDI event from the circular input buffer.  Events
 *  are removed from the buffer.  The corresponding PutEvent() function
 *  is called by the low-level callback function, so it must reside in
 *  the callback DLL.
 *
 * Params:  bufflist - Points to the circular buffer.
 *          lpEvent - Points to an EVENT structure that is filled with the
 *              retrieved event.
 *
 * Return:  Returns non-zero if successful, zero if there are no 
 *   events to get.
 */

static WORD FAR PASCAL
MidiGetEvent(LPBufferList bufflist, LPEVENT lpEvent)
{
    /* If no event available, return.
     */
restart:
    if(bufflist->sofar <= 0)
        return 0;
    
	// For the moment, we handle the output buffer release events 
	// in here, though they should eventually be sent to the output thread.

	if ( bufflist->tail->msg == WM_MIDIOUTPUT ) {

		handlemidioutput(bufflist->tail->data);

		// bogus code, will go away when it moves to output thread
    	--bufflist->sofar;
    	++bufflist->tail;
    	if(bufflist->tail >= bufflist->endoflist)
    	    bufflist->tail = bufflist->list;
		goto restart;
	}

    /* Get the event.
     */
    *lpEvent = *bufflist->tail;

    /* Decrement the byte count, bump the tail pointer.
     */
    --bufflist->sofar;
    ++bufflist->tail;
    
    /* Wrap the tail pointer, if necessary.
     */
    if(bufflist->tail >= bufflist->endoflist)
        bufflist->tail = bufflist->list;

    return 1;
}

/* AllocCallbackInstanceData - Allocates a MidiCallback
 *      structure.  This structure is used to pass information to the
 *      low-level callback function, each time it receives a message.
 *
 * Params:  void
 *
 * Return:  A pointer to the allocated CALLBACKINSTANCE data structure.
 */

static LPMidiCallback FAR PASCAL
AllocCallbackInstanceData()
{
    LPMidiCallback bufflist;
    
    bufflist = (LPMidiCallback) GlobalAlloc(0, (DWORD)sizeof(MidiCallback));
    return bufflist;
}

/* FreeCallbackInstanceData - Frees the given MidiCallback structure.
 *
 * Params:  bufflist - Points to the MidiCallback structure to be freed.
 *
 * Return:  void
 */

static void FAR PASCAL
FreeCallbackInstanceData(LPMidiCallback bufflist)
{
    GlobalFree(bufflist);
}

static void
midiinerror(JNIEnv *env, int r, char *s)
{
	char ebuff[EBUFFSIZE];
	int lng;

	strcpy(ebuff,s);
	if ( r >= 0 ) {
		lng = strlen(ebuff);
		midiInGetErrorText(r, ebuff+lng, EBUFFSIZE-lng);
	}
	// printf(ebuff);
	throwMidiException(env,ebuff);
}

static void
midiouterror(JNIEnv *env, int r, char *s)
{
	char ebuff[EBUFFSIZE];
	int lng;

	strcpy(ebuff,s);
	if ( r >= 0 ) {
		lng = strlen(ebuff);
		midiOutGetErrorText(r, ebuff+lng, EBUFFSIZE-lng);
	}
	throwMidiException(env,ebuff);
}

void FAR PASCAL midiInputHandler(
	HMIDIIN handle_in, 
	WORD mtype, 		/* MIM_* values */
	DWORD cbinstance, 	/* callback instance */
	DWORD dwParam1, 	/* MIDI data */
	DWORD dwParam2)		/* milliseconds */
{
	static EVENT event;

	switch(mtype) {
        case MIM_DATA:
		if ( LOBYTE(LOWORD(dwParam1)) == (BYTE)0xfe )
			break;
		/* continue on */

	case MIM_LONGDATA:

		event.device = ((LPMidiCallback)cbinstance)->device;
		event.data = dwParam1;		/* real MIDI data */
		event.timestamp = dwParam2;
		event.islong = (mtype==MIM_LONGDATA);
		event.msg = WM_MIDIINPUT;

		MidiPutEvent(((LPMidiCallback)cbinstance)->bufflist,
			(LPEVENT) &event); 

		SetEvent(midiInEvent);

		break;
	case MIM_OPEN:
		break;
	case MIM_CLOSE:
		break;
	case MIM_ERROR:
	case MIM_LONGERROR:

		event.device = ((LPMidiCallback)cbinstance)->device;
		event.data = dwParam1;
		event.timestamp = dwParam2;
		event.msg = WM_MIDI_ERROR;
		event.islong = 0;

		MidiPutEvent(((LPMidiCallback)cbinstance)->bufflist,
			(LPEVENT) &event); 

		SetEvent(midiInEvent);
		break;
    default:
		break;
	}
}

void FAR PASCAL midiOutputHandler(
	HMIDIIN handle_out,
	WORD mtype,		/* MOM_* values */
	DWORD cbinstance, 	/* callback instance */
	DWORD dwParam1, 	/* MIDI data */
	DWORD dwParam2)		/* milliseconds */
{
	static EVENT event;

	switch(mtype) {
	case MOM_OPEN:
	case MOM_CLOSE:
		break;

	case MOM_DONE:
		event.device = ((LPMidiCallback)cbinstance)->device;
		event.data = dwParam1;
		event.timestamp = dwParam2;
		event.msg = WM_MIDIOUTPUT;
		event.islong = 0;

		MidiPutEvent(((LPMidiCallback)cbinstance)->bufflist, (LPEVENT)&event); 
		SetEvent(midiInEvent);	/* re-use input event queue */
		break;

	default:
		break;
    }
}

void FAR PASCAL MidiPutEvent(LPBufferList buff, LPEVENT ev)
{
	if(buff->sofar >= buff->total)
		return;
    
	*buff->head = *ev;
	(buff->head)++;
	(buff->sofar)++;

	if ( buff->head >= buff->endoflist )
		buff->head = buff->list;
}

#include <ddeml.h>

#ifdef UNICODE
#define STRICMP wcsicmp
#define ITOA(c, sz, b) (itoa(sizeof(szA), szA, b), mbstowcs(sz, szA, b), sz)
#else
#define STRICMP stricmp
#define ITOA itoa
#endif  // UNICODE

#define DDETIMEOUT 5000

// TCHAR szNetscape[] = TEXT("IEXPLORE"); // DDE service name
TCHAR szNetscape[] = TEXT("NETSCAPE"); // DDE service name

TCHAR szWWW_GetWindowInfo[] = TEXT("WWW_GetWindowInfo");    // DDE topic name
TCHAR szWWW_OpenURL[] = TEXT("WWW_OpenURL");    // DDE topic name
TCHAR szWindowID[] = TEXT("dwWindowID");    // DDE topic name

HSZ         hszBrowser = 0;
HSZ         hszWWW_GetWindowInfo = 0;
HSZ         hszWWW_OpenURL = 0;
HSZ         hszWindowID = 0;
HSZ         hszBuff = 0;

DWORD       idInst = 0;                 // our DDEML instance object
HCONV		hConv = 0;              // the list of all convs we have open
HDDEDATA	hRet;
#ifdef UNICODE
CHAR        szA[20];                    // static buffer for UNICODE conversion
#endif

JNIEXPORT void JNICALL
Java_com_nosuch_midi_MidiSystem_nativeOpenURL(JNIEnv *env, jobject obj, jstring browser, jstring url)
{

	char buff[256];
	char *pdata;

	if (DdeInitialize(&idInst, (PFNCALLBACK)0,
		APPCMD_FILTERINITS | APPCMD_CLIENTONLY, 0)) {
		throwMidiException(env,"Unable to initialize DDE in nativeOpenURL");
		return;
	}

	/*
	 * Initialize all our string handles for use later
	 */
	hszBrowser = DdeCreateStringHandle(idInst, szNetscape, 0);
	hszWWW_GetWindowInfo = DdeCreateStringHandle(idInst,szWWW_GetWindowInfo,0);
	hszWWW_OpenURL = DdeCreateStringHandle(idInst,szWWW_OpenURL,0);
	hszWindowID = DdeCreateStringHandle(idInst, szWindowID, 0);

#ifdef OLDSTUFF
	// Start the conversation to get the URL in the current window
	hConv = DdeConnect(idInst, hszBrowser, hszWWW_GetWindowInfo, NULL);
	strcpy(buff,"0xFFFFFFFF");
	hszBuff = DdeCreateStringHandle(idInst, buff, 0);
	hRet = DdeClientTransaction(NULL, 0ul, hConv, hszBuff,
		 CF_TEXT, XTYP_REQUEST, DDETIMEOUT, NULL);
	pdata = (char *)DdeAccessData(hRet, NULL);
	DdeDisconnect(hConv);
#endif

	// Start the conversation to open a URL in the current window
	hConv = DdeConnect(idInst, hszBrowser, hszWWW_OpenURL, NULL);
	strcpy(buff,"http://www.nosuch.com/cook/cookget.cgi?id=333,,0xFFFFFFFF,0,,,");
	hszBuff = DdeCreateStringHandle(idInst, buff, 0);
	hRet = DdeClientTransaction(NULL, 0ul, hConv, hszBuff,
		 CF_TEXT, XTYP_REQUEST, DDETIMEOUT, NULL);
	pdata = (char *)DdeAccessData(hRet, NULL);

	DdeDisconnect(hConv);

	DdeFreeStringHandle(idInst, hszBrowser);
	DdeFreeStringHandle(idInst, hszWWW_GetWindowInfo);

	DdeUninitialize(idInst);

	return;
}
